/*
 * isnsdd - the iSNS Discovery Daemon
 *
 * Copyright (C) 2007 Olaf Kirch <olaf.kirch@oracle.com>
 */

#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>

#ifdef MTRACE
# include <mcheck.h>
#endif

#include <isns.h>
#include "security.h"
#include "util.h"
#include "isns-proto.h"
#include "paths.h"
#include "attrs.h"

enum {
	ROLE_INITIATOR = 1,
	ROLE_MONITOR = 2,
};

#define MAX_RETRY_TIMEOUT	300

static const char *	opt_configfile = ISNS_DEFAULT_ISNSDD_CONFIG;
static int		opt_af = AF_INET6;
static int		opt_foreground = 0;
static int		opt_role = ROLE_INITIATOR;
static int		opt_scn_bits = ISNS_SCN_OBJECT_UPDATED_MASK |
				ISNS_SCN_OBJECT_ADDED_MASK |
				ISNS_SCN_OBJECT_REMOVED_MASK |
				ISNS_SCN_TARGET_AND_SELF_ONLY_MASK;
static unsigned int	opt_retry_timeout = 10;
static int		opt_esi = 1;

static isns_socket_t *	server_socket;
static isns_db_t *	my_db;
static char *		my_entity_id;
static isns_object_list_t my_objects = ISNS_OBJECT_LIST_INIT;
static time_t		my_last_registration;
static unsigned int	esi_interval;

static void		run_discovery(isns_server_t *srv);
static void		scn_callback(isns_db_t *, uint32_t,
				isns_object_template_t *,
				const char *);
static void		refresh_registration(void *);
static void		retry_registration(void *);
static void		zap_objects(isns_object_list_t *);
static void		usage(int, const char *);
static void		cleanup(int);

static struct option	options[] = {
      {	"config",		required_argument,	NULL,	'c'		},
      {	"debug",		required_argument,	NULL,	'd'		},
      { "foreground",		no_argument,		NULL,	'f'		},
      { "role",			required_argument,	NULL,	'r'		},
      { "no-esi",		no_argument,		NULL,	'E'		},
      { "help",			no_argument,		NULL,	'h'		},
      { "version",		no_argument,		NULL,	'V'		},
      { NULL }
};

int
main(int argc, char **argv)
{
	isns_server_t	*server;
	isns_source_t	*source;
	int		c;

#ifdef MTRACE
	mtrace();
#endif

	while ((c = getopt_long(argc, argv, "46c:d:Efhr:", options, NULL)) != -1) {
		switch (c) {
		case '4':
			opt_af = AF_INET;
			break;

		case '6':
			opt_af = AF_INET6;
			break;

		case 'c':
			opt_configfile = optarg;
			break;

		case 'd':
			isns_enable_debugging(optarg);
			break;

		case 'E':
			opt_esi = 0;
			break;

		case 'f':
			opt_foreground = 1;
			break;

		case 'h':
			usage(0, NULL);

		case 'r':
			if (!strcasecmp(optarg, "initiator"))
				opt_role = ROLE_INITIATOR;
			else
			if (!strcasecmp(optarg, "control")
			 || !strcasecmp(optarg, "monitor"))
				opt_role = ROLE_MONITOR;
			else {
				isns_error("Unknown role \"%s\"\n", optarg);
				usage(1, NULL);
			}
			break;

		case 'V':
			printf("Open-iSNS version %s\n"
			       "Copyright (C) 2007, Olaf Kirch <olaf.kirch@oracle.com>\n",
			       OPENISNS_VERSION_STRING);
			return 0;

		default:
			usage(1, "Unknown option");
		}
	}

	if (optind != argc)
		usage(1, NULL);

	isns_read_config(opt_configfile);

	if (!isns_config.ic_source_name)
		usage(1, "Please specify an iSNS source name");
	source = isns_source_create_iscsi(isns_config.ic_source_name);

	isns_write_pidfile(isns_config.ic_pidfile);

	/* Create an in-memory database for now. */
	my_db = isns_db_open(NULL);
	if (my_db == NULL)
		isns_fatal("Unable to open database\n");

	if (!opt_foreground) {
		if (daemon(0, 0) < 0)
			isns_fatal("Unable to background server process\n");
		isns_log_background();
		isns_update_pidfile(isns_config.ic_pidfile);
	}

	signal(SIGTERM, cleanup);
	signal(SIGINT, cleanup);

	server = isns_create_server(source, my_db, &isns_callback_service_ops);
	isns_server_set_scn_callback(server, scn_callback);

	run_discovery(server);
	return 0;
}

void
usage(int exval, const char *msg)
{
	if (msg)
		fprintf(stderr, "Error: %s\n", msg);
	fprintf(stderr,
	"Usage: isnsdd [options]\n\n"
	"  --role <role>   Specify role (one of initiator, control)\n"
	"  --config        Specify alternative config fille\n"
	"  --foreground    Do not put daemon in the background\n"
	"  --no-esi        Do not try to register an portals for ESI status inquiries\n"
	"  --debug         Enable debugging (list of debug flags)\n"
	"  --help          Print this message\n"
	);
	exit(exval);
}

void
cleanup(int sig)
{
	isns_remove_pidfile(isns_config.ic_pidfile);
	exit(1);
}

/*
 * Force a re-registration of the whole object set.
 */
static void
force_reregistration(isns_client_t *clnt)
{
	isns_cancel_timer(refresh_registration, clnt);
	isns_add_oneshot_timer(0, retry_registration, clnt);
}

/*
 * Refresh the registration by calling DevAttrQry
 */
static void
refresh_registration(void *ptr)
{
	isns_object_list_t objects = ISNS_OBJECT_LIST_INIT;
	isns_client_t	*clnt = (isns_client_t *) ptr;
	isns_attr_list_t query_key = ISNS_ATTR_LIST_INIT;
	isns_simple_t	*qry = NULL;
	uint32_t	status;

	isns_debug_state("Refreshing registration\n");
	isns_attr_list_append_string(&query_key,
			ISNS_TAG_ENTITY_IDENTIFIER,
			my_entity_id);

	qry = isns_create_query(clnt, &query_key);
	isns_attr_list_destroy(&query_key);

	/* We should have an async call mechanism. If the server
	 * is wedged, we'll block here, unable to service any other
	 * functions.
	 */
	status = isns_simple_call(clnt->ic_socket, &qry);
	if (status != ISNS_SUCCESS) {
		isns_error("Query failed: %s\n", isns_strerror(status));
		goto re_register;
	}

	status = isns_query_response_get_objects(qry, &objects);
	if (status != ISNS_SUCCESS) {
		isns_error("Unable to parse query response\n");
		goto re_register;
	}

	if (objects.iol_count == 0)
		goto re_register;

out:
	isns_simple_free(qry);
	return;

re_register:
	isns_warning("Lost registration, trying to re-register\n");
	force_reregistration(clnt);
	goto out;
}

/*
 * Check if all portals have seen ESI messages from the server
 */
static void
check_portal_registration(void *ptr)
{
	isns_object_list_t bad_portals = ISNS_OBJECT_LIST_INIT;
	isns_client_t	*clnt = (isns_client_t *) ptr;
	unsigned int	i, need_reregister = 0, good_portals = 0;
	time_t		now;

	isns_debug_state("%s()\n", __FUNCTION__);
	now = time(NULL);
	for (i = 0; i < my_objects.iol_count; ++i) {
		isns_object_t *obj = my_objects.iol_data[i];
		isns_portal_info_t portal_info;
		time_t		last_modified;
		uint32_t	interval;

		if (!isns_object_is_portal(obj))
			continue;
		if (!isns_object_get_uint32(obj, ISNS_TAG_ESI_INTERVAL, &interval))
			continue;

		last_modified = isns_object_last_modified(obj);
		if (last_modified + 2 * interval > now) {
			good_portals++;
			continue;
		}

		isns_portal_from_object(&portal_info,
				ISNS_TAG_PORTAL_IP_ADDRESS,
				ISNS_TAG_PORTAL_TCP_UDP_PORT,
				obj);

		isns_notice("Portal %s did not receive ESIs within %u seconds - "
			"server may have lost us.\n",
			isns_portal_string(&portal_info),
			now - last_modified);

		/* If we haven't received ANY ESIs, ever, the portal
		 * may be using a private IP */
		if (last_modified <= my_last_registration)
			isns_object_list_append(&bad_portals, obj);

		need_reregister++;
	}

	if (bad_portals.iol_count == my_objects.iol_count) {
		isns_notice("Suspiciously little ESI traffic - server may be broken\n");
		/* Delete our node and portals; and force a
		 * re-registration from scratch. This time without ESI.
		 */
		zap_objects(&my_objects);
		opt_esi = 0;
	}

	for (i = 0; i < bad_portals.iol_count; ++i)
		isns_object_list_remove(&my_objects, bad_portals.iol_data[i]);
	zap_objects(&bad_portals);

	if (need_reregister)
		force_reregistration(clnt);
}

/*
 * Zap the list of objects, and remove them from our
 * shadow DB
 */
static void
zap_objects(isns_object_list_t *list)
{
	unsigned int	i;

	for (i = 0; i < list->iol_count; ++i) {
		isns_object_t *obj = list->iol_data[i];

		isns_db_remove(my_db, obj);
	}
	isns_object_list_destroy(list);
	isns_db_purge(my_db);
}

/*
 * Given the DevAttrReg response, extract the entity ID we
 * have been assigned.
 */
static int
extract_entity_id(isns_client_t *clnt, isns_simple_t *resp)
{
	isns_object_t	*entity = NULL;
	int		status;
	unsigned int	i;

	zap_objects(&my_objects);

	status = isns_query_response_get_objects(resp, &my_objects);
	if (status) {
		isns_error("Unable to extract object list from "
			   "registration response: %s\n",
			   isns_strerror(status), status);
		return status;
	}

	for (i = 0; i < my_objects.iol_count; ++i) {
		isns_object_t	*obj = my_objects.iol_data[i];
		uint32_t	interval;
		const char	*eid;

		if (!isns_object_is_entity(obj))
			continue;

		if (entity) {
			isns_error("Server returns more than one entity "
				   "in registration response. What a weirdo.\n");
			continue;
		}

		entity = obj;
		if (!isns_object_get_string(obj,
					ISNS_TAG_ENTITY_IDENTIFIER,
					&eid)) {
			isns_error("Unable to obtain EID from entity!\n");
			status = ISNS_NO_SUCH_ENTRY;
			break;
		}

		isns_assign_string(&my_entity_id, eid);
		if (isns_object_get_uint32(obj,
					ISNS_TAG_REGISTRATION_PERIOD,
					&interval)) {
			if (interval == 0) {
				isns_error("Server returns a registration period of 0\n");
			} else {
				isns_debug_state("Setting up timer for registration refresh\n");
				isns_add_timer(interval / 2,
						refresh_registration,
						clnt);
			}
		}
		break;
	}

	isns_cancel_timer(check_portal_registration, clnt);
	esi_interval = 0;

	for (i = 0; i < my_objects.iol_count; ++i) {
		isns_object_t	*obj = my_objects.iol_data[i];
		uint32_t	interval;

		/* Attach all portals, nodes etc to the entity */
		if (entity && obj != entity)
			isns_object_attach(obj, entity);

		/* Insert everything but portal groups */
		if (!isns_object_is_pg(obj))
			isns_db_insert(my_db, obj);

		if (!isns_object_is_portal(obj))
			continue;

		if (isns_object_get_uint32(obj, ISNS_TAG_ESI_INTERVAL, &interval)) {
			if (interval == 0) {
				isns_error("Server returns an ESI interval of 0\n");
			} else if (!esi_interval || interval < esi_interval) {
				esi_interval = interval;
			}
		}
	}

	if (esi_interval) {
		isns_debug_state("Setting up timer to check for ESI reachability\n");
		isns_add_timer(esi_interval * 4 / 5,
				check_portal_registration,
				clnt);
	}

	my_last_registration = time(NULL);
	return status;
}

static int
build_objects(isns_source_t *source,
		uint32_t node_type,
		isns_portal_info_t *portal_info,
		int with_esi,
		isns_object_list_t *objects)
{
	isns_object_t	*node, *entity;
	unsigned int	i, nportals = 1;

	/* If our service socket is bound to the wilcard
	   address, portal_info->addr will be IN6ADDR_ANY,
	   which is hardly something we want to register. */
	if (isns_portal_is_wildcard(portal_info)) {
		static isns_portal_info_t iflist[16];

		nportals = isns_enumerate_portals(iflist, 16);
		if (nportals == 0)
			isns_fatal("Unable to enumerate portals - "
				   "no usable interfaces found\n");

		for (i = 0; i < nportals; ++i) {
			iflist[i].addr.sin6_port = portal_info->addr.sin6_port;
			iflist[i].proto = portal_info->proto;
		}
		portal_info = iflist;
	}

	entity = NULL;
	node = isns_create_storage_node2(source, node_type, entity);
	isns_object_list_append(objects, node);

	for (i = 0; i < nportals; ++i, ++portal_info) {
		isns_object_t	*portal, *pg;

		portal = isns_create_portal(portal_info, entity);
		isns_object_list_append(objects, portal);

		/* Always put initiator node and portal into a group
		 * of their own. We reserve PG tag 2 for this. */
		pg = isns_create_portal_group(portal, node, 2);
		isns_object_list_append(objects, pg);

		isns_object_set_uint32(portal,
				ISNS_TAG_SCN_PORT,
				isns_portal_tcpudp_port(portal_info));
		if (with_esi) {
			isns_object_set_uint32(portal,
					ISNS_TAG_ESI_INTERVAL,
					isns_config.ic_esi_min_interval);
			isns_object_set_uint32(portal,
					ISNS_TAG_ESI_PORT,
					isns_portal_tcpudp_port(portal_info));
		}
	}

	return 0;
}

static int
register_objects(isns_client_t *clnt, isns_object_list_t *objects)
{
	isns_object_t	*entity = NULL;
	isns_simple_t	*reg;
	unsigned int	i;
	int		status;

	for (i = 0; i < objects->iol_count; ++i) {
		if (isns_object_is_entity(objects->iol_data[i])) {
			entity = objects->iol_data[i];
			break;
		}
	}

	reg = isns_create_registration(clnt, entity);
	for (i = 0; i < objects->iol_count; ++i)
		isns_registration_add_object(reg, objects->iol_data[i]);
	isns_registration_set_replace(reg, 1);

	status = isns_simple_call(clnt->ic_socket, &reg);
	if (status != ISNS_SUCCESS)
		return status;

	/* Extract the EID and registration period */
	extract_entity_id(clnt, reg);
	isns_simple_free(reg);

	reg = isns_create_scn_registration(clnt, opt_scn_bits);
	status = isns_simple_call(clnt->ic_socket, &reg);

	if (status != ISNS_SUCCESS) {
		isns_error("SCN registration failed: %s\n", isns_strerror(status));
		isns_cancel_timer(refresh_registration, clnt);
	}

	isns_simple_free(reg);
	return status;
}

static int
scn_register(isns_client_t *clnt)
{
	isns_portal_info_t portal_info;
	int		status, with_esi, node_type;

	if (!isns_socket_get_portal_info(server_socket, &portal_info))
		isns_fatal("Unable to get portal info\n");

	switch (opt_role) {
	case ROLE_INITIATOR:
		node_type = ISNS_ISCSI_INITIATOR_MASK;
		break;

	case ROLE_MONITOR:
		node_type = ISNS_ISCSI_CONTROL_MASK;
		break;

	default:
		isns_fatal("Bad role <%u>\n", opt_role);
		return 0; /* compiler food */
	}

	/* Try ESI first. If the server doesn't support it, or doesn't
	 * have the resources to serve us, fall back to normal
	 * registration refresh. */
	with_esi = opt_esi;
	while (1) {
		build_objects(clnt->ic_source, node_type,
				&portal_info, with_esi, &my_objects);

		status = register_objects(clnt, &my_objects);
		if (status == ISNS_SUCCESS)
			break;

		isns_object_list_destroy(&my_objects);

		if (status == ISNS_ESI_NOT_AVAILABLE && with_esi--)
			continue;

		isns_error("Failed to register object(s): %s\n",
				isns_strerror(status));
		return status;
	}

	return status;
}

static void
retry_registration(void *ptr)
{
	isns_client_t *clnt = (isns_client_t *) ptr;
	static unsigned int timeout = 0;
	int	status;

	if (my_objects.iol_count == 0) {
		/* Something bad happened; all objects were zapped.
		 * Start from scratch
		 */
		status = scn_register(clnt);
	} else {
		status = register_objects(clnt, &my_objects);
	}

	if (status) {
		if (timeout == 0)
			timeout = opt_retry_timeout;
		else if (timeout >= MAX_RETRY_TIMEOUT)
			timeout = MAX_RETRY_TIMEOUT;

		isns_debug_state("Retrying to register in %u seconds\n", timeout);
		isns_add_oneshot_timer(timeout, retry_registration, clnt);

		/* Exponential backoff */
		timeout <<= 1;
	}
}

/*
 * Invoke the registered callout program
 */
static void
callout(const char *how, isns_object_t *obj, unsigned int bitmap)
{
	char	*argv[128];
	int	fargc, argc = 0;
	pid_t	pid;

	if (!isns_config.ic_scn_callout)
		return;

	argv[argc++] = isns_config.ic_scn_callout;
	argv[argc++] = (char *) how;
	fargc = argc;

	argc += isns_print_attrs(obj, argv + argc, 128 - argc);

	pid = fork();
	if (pid == 0) {
		execv(argv[0], argv);
		isns_fatal("Cannot execute %s: %m\n", argv[0]);
	}

	while (fargc < argc)
		isns_free(argv[fargc++]);

	if (pid < 0) {
		isns_error("fork: %m\n");
		return;
	}

	while (waitpid(pid, NULL, 0) < 0)
		;
}

/*
 * This is called when we receive a State Change Notification
 */
static void
scn_callback(isns_db_t *db, uint32_t bitmap,
		isns_object_template_t *node_type,
		const char *node_name)
{
	isns_object_t	*node;
	uint32_t	tag;

	isns_notice("%s \"%s\" %s\n",
			isns_object_template_name(node_type),
			node_name, isns_event_string(bitmap));

	/* FIXME: We should update our own shadow DB here, and
	 * possibly trigger callouts. */
	if (node_type == &isns_iscsi_node_template) {
		tag = ISNS_TAG_ISCSI_NAME;
	} else
#if 0
	if (node_type == &isns_fc_node_type) {
		tag = ISNS_TAG_FC_PORT_NAME_WWPN;
	} else
#endif
		return;

	node = isns_db_vlookup(db, node_type, tag, node_name, 0);
	if (bitmap & (ISNS_SCN_OBJECT_REMOVED_MASK|ISNS_SCN_DD_MEMBER_REMOVED_MASK)) {
		if (node) {
			isns_db_remove(db, node);
			isns_db_purge(db);
		}
		callout("remove", node, bitmap);
	} else
	if (bitmap & (ISNS_SCN_OBJECT_ADDED_MASK|ISNS_SCN_OBJECT_UPDATED_MASK|ISNS_SCN_DD_MEMBER_ADDED_MASK)) {
		const char	*how = "add";

		if (bitmap & ISNS_SCN_OBJECT_UPDATED_MASK)
			how = "update";
		if (!node) {
			node = isns_create_object(node_type, NULL, NULL);
			if (!node)
				return;
			isns_object_set_string(node, tag, node_name);
			isns_db_insert(db, node);
		}
		callout(how, node, bitmap);

		/* Query the node and get updated portals/PGs */
	}
}

/*
 * Server main loop
 */
void
run_discovery(isns_server_t *server)
{
	isns_client_t	*clnt;
	isns_security_t	*ctx = NULL;
	isns_message_t	*msg, *resp;

	ctx = isns_default_security_context(0);

#if 0
	if (isns_config.ic_control_socket) {
		server_socket = isns_create_server_socket(isns_config.ic_control_socket,
				NULL, AF_UNSPEC, SOCK_STREAM);
		if (server_socket == NULL)
			isns_fatal("Unable to create control socket\n");
		isns_socket_set_security_ctx(server_socket, ctx);
	}
#endif

	/* Create the SCN socket */
	server_socket = isns_create_server_socket(isns_config.ic_bind_address,
			NULL, opt_af, SOCK_DGRAM);
	if (server_socket == NULL)
		isns_fatal("Unable to create server socket\n");
	isns_socket_set_security_ctx(server_socket, ctx);

	clnt = isns_create_default_client(NULL);
	if (clnt == NULL)
		isns_fatal("Cannot connect to server\n");

	if (scn_register(clnt))
		isns_fatal("Initial registration failed - giving up.\n");

	while (1) {
		struct timeval timeout = { 0, 0 };
		time_t	now, then, next_timeout;
		unsigned int function;

		next_timeout = time(NULL) + 3600;

		/* Run timers */
		then = isns_run_timers();
		if (then && then < next_timeout)
			next_timeout = then;

		/* Determine how long we can sleep */
		now = time(NULL);
		if (next_timeout <= now)
			continue;
		timeout.tv_sec = next_timeout - now;

		if ((msg = isns_recv_message(&timeout)) == NULL)
			continue;

		function = isns_message_function(msg);
		if (function != ISNS_STATE_CHANGE_NOTIFICATION
		 && function != ISNS_ENTITY_STATUS_INQUIRY) {
			isns_warning("Discarding unexpected %s message\n",
					isns_function_name(function));
			isns_message_release(msg);
			continue;
		}

		if ((resp = isns_process_message(server, msg)) != NULL) {
			isns_socket_t *sock = isns_message_socket(msg);

			isns_socket_send(sock, resp);
			isns_message_release(resp);
		}

		isns_message_release(msg);
	}
}
